import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import '../../bloc/bloc.dart';
import '../../models/graph_node.dart';
import 'stage_entity.dart';
import '../../mixin/selectable.dart';

class WrapperBox extends StageEntity {

  final GraphNode node;

  WrapperBox({
    Key key,
    @required this.node,
  }) : super(key: key);

  @override
  WrapperBoxState createState() => WrapperBoxState();
}

class WrapperBoxState extends StageEntityState<WrapperBox> with Selectable {

  GraphNode _node;

  bool _hovering = false;

  /// gives false positive..
  /// ignore: unused_field
  bool _dragging = false;

  GraphNode get node => _node;

  @override void initState() {
    super.initState();

    _node = widget.node;
    selected = false;

    String selectedID = stageBloc.currentState?.selectedBox?._node?.id ?? "-1";

    if (_node.id == selectedID) {
      stageBloc.setSelectedBox(this);
    }

    WidgetsBinding.instance.addPostFrameCallback(_onBuildCompleted);
  }

  void _onBuildCompleted(_) {
    if (_node.size == null || _node.size == Size.zero) {
      Size containerSize = _getContainerSize();
      _node.size = Size(containerSize.width, containerSize.height);
    }
  }

  Size _getContainerSize() {
    final RenderBox containerRenderBox = context.findRenderObject() as RenderBox;
    return containerRenderBox.size ?? Size.zero;
  }

  @override
  void onSelected() {
    updateOnScale(stageBloc.currentState.stageScale);
    super.onSelected();
  }


  @override 
  void updateOnScale(double newScale) {
    stageScale = newScale;
    /// stageInverseScale = stageScale < 1.0 ? (1.0 / stageScale) : 1.0;
    stageInverseScale = stageScale > 0.0 ? (1.0 / stageScale) : 1.0;
  }

  @override
  void onStageTransformChange(Matrix4 transform) {

    if (!mounted) return;

    /// only redraw, if selected
    if (selected) {
      double newScale = transform.getMaxScaleOnAxis();
      if (stageScale != newScale) {
        startAnim(stageScale, newScale);
      }
    }
  }

  void _onTapUp(TapUpDetails details) {
    stageBloc.setSelectedBox(this);
  }

  Offset get centre => Offset(_node.position.dx + _node.size.width / 2.0, _node.position.dy + _node.size.height / 2.0);

  Rect asRect() => Rect.fromLTWH(_node.position.dx, _node.position.dy, _node.size.width, _node.size.height);

  // prevent scroll out from the area
  Offset _checkBounds(Offset delta, Rect bounds) {
    
    Rect box = asRect();
    box = box.translate(delta.dx, delta.dy);
    double left = _node.position.dx;
    double top = _node.position.dy;

    if (box.left < bounds.left)
      left = bounds.left;
    else
    if (box.right > bounds.right)
      left = bounds.right - _node.size.width;

    if (box.top < bounds.top)
      top = bounds.top;
    else
    if (box.bottom > bounds.bottom)
      top = bounds.bottom - _node.size.height;

    return Offset(left, top);
  }

  void handlePanStart(DragStartDetails details) {
    setState(() {
      /// _node.position = details.globalPosition;
    });
  }

  void handlePanUpdate(DragUpdateDetails details) {
    /*
    Matrix4 transform = StageSessionData.currentState.stageTransform;
    double scale = transform.getMaxScaleOnAxis();
    */
    /// RenderBox box = this.context.findRenderObject() as RenderBox;
     
    Offset delta = details.delta; //  * stageInverseScale;
 
    Offset newPos = _checkBounds(delta, stageBloc.currentState.bounds);
 
    if (!mounted) return;
    setState(() {
      _node.position = newPos + delta;
    });
     
    graphBloc.notifyConnectedEdges(_node);
  }

  void handlePanEnd(DragEndDetails details) {

    Offset newPos = _checkBounds(Offset.zero, stageBloc.currentState.bounds);

    if (!mounted) return;
    setState(() {
      _node.position = newPos;
    });
  }

  void handleDoubleTap() {
    //print("double tap");
    this.stageBloc.setSelectedBox(this);
    this.stageBloc.setConnectionMode(true);
  }

  Widget build(BuildContext context) {
  
    final bool inConnectMode = stageBloc.currentState.connectionEnabled;

    bool amIDraggable = selected && inConnectMode;
    bool amIDragtarget = !selected && inConnectMode;

    return Positioned(
      left: _node.position.dx,
      top: _node.position.dy,
      child: amIDraggable ? 
          _iAmDraggable()
        :  GestureDetector(
            onTapUp: _onTapUp,
            onDoubleTap: handleDoubleTap,
            ///onPanStart: selected ? handlePanStart : null,
            onPanUpdate: selected ? handlePanUpdate : null,
            onPanEnd: selected ? handlePanEnd : null,
            child: amIDragtarget ? 
              _iAmDragTarget()
            : _getBody(borderColor: stageBloc.theme.boxSelectedOutLineColor, inset: selected)
          )
    );
  }

  Widget _iAmDraggable() {
    return Draggable(
      data: _node.id, // WrapperBoxState
      feedback: Transform.scale(
        scale: stageScale,
        child: Opacity(
          opacity: 0.75, //_opacity,
          /// duration: Duration(milliseconds: 300),
          child: _getBody(borderColor: stageBloc.theme.boxSelectedOutLineColor, inset: selected)
        ),
      ),
      onDragStarted: () {
        if (!mounted) return;
        setState(() {
          this._dragging = true; 
        });
      },
      onDragEnd: (DraggableDetails details) {
        if (!mounted) return;
          setState(() {
            this._dragging = false; 
          });
      },
      childWhenDragging: _getBody(borderColor: stageBloc.theme.boxOutlineDragColor),
      child: _getBody(borderColor: stageBloc.theme.boxSelectedOutLineColor) //childDrag
    
    );
  }

  Widget _iAmDragTarget() {
    
    return DragTarget(
      builder: (context, dataToAccept, dataToReject) {
        return _hovering ? _getBody(borderColor: stageBloc.theme.boxHoverColor)
         : _getBody(borderColor: stageBloc.theme.boxSelectedOutLineColor, inset: false);
      },
      onLeave: (String senderNodeID) {
        if (!mounted) return;
        setState(() {
         _hovering = false; 
        });
      },
      onWillAccept: (String senderNodeID) {
        if (!mounted) return false;
        setState(() {
          _hovering = true;
        });
        return true;
      },
      // onAccept is only called, if onWillAccept returns true 
      onAccept: (String senderNodeID) {
        if (!mounted) return;
        setState(() {
         _hovering = false; 
        });

        graphBloc.dispatch( ConnectNodesEvent(senderNodeID, _node.id) );
      }
    );
  }

  Widget _getBody({@required Color borderColor, bool inset = true}) {
    final double borderWidth = 3.0 + (2.0 * stageInverseScale);
    return ClipRRect(
      borderRadius: BorderRadius.circular(0),
      child: Container(
        width: _node.size.width == 0 ? null : _node.size.width, 
        height: _node.size.height == 0 ? null : _node.size.height, 
        decoration: BoxDecoration(
          boxShadow: [
            BoxShadow(
              color: Color.fromARGB(60, 0, 0, 0), 
              blurRadius: 8
            )
          ],
          border: inset ? 
                  Border.all(
                    style: BorderStyle.solid, 
                    color: borderColor, 
                    width: borderWidth
                  )
                : null,
        ),
        ///padding: inset ? EdgeInsets.all(borderWidth) : null, 
        child: _node.widget ?? null
      )
    );
  }

  @override void dispose() {
    super.dispose();   
  }
}
